/* jshint esversion : 6 */

(function(w, d, undefined) {

    function polyfill() {
        // return when scrollBehavior interface is supported
        if ('scrollBehavior' in d.documentElement.style)
            return;


        // globals 
        const Element = w.HTMLElement || w.Element,
            SCROLL_TIME = 468;


        /* object gathering original scroll methods */
        let original = {
            scroll: w.scroll || w.scrollTo,
            scrollBy: w.scrollBy,
            elScroll: Element.prototype.scroll || scrollElement,
            scrollIntoView: Element.prototype.scrollIntoView
        };


        /* define timing method */
        let now = w.performance && w.performance.now ?
            w.performance.now.bind(w.performance) : Date.now;

        /*
         * changes scroll position inside an element
         * @method scrollElement
         * @param {Number} x
         * @param {Number} y
         */
        function scrollElement(x, y) {
            this.scrollLeft = x;
            this.scrollTop = y;
        }

        /*
         * returns result of applying ease math function to a number
         * @method ease
         * @param {Number} k
         * @returns {Number}
         */
        function ease(k) {
            return 0.5 * (1 - Math.cos(Math.PI * k));
        }

        /*
         * indicates if a smooth behavior should be applied
         * @method shouldBailOut
         * @param {Number|Object} x
         * @returns {Boolean}
         */
        function shouldBailOut(x) {
            if (typeof x !== 'object' ||
                x === null ||
                x.behavior === undefined ||
                x.behavior === 'auto' ||
                x.behavior === 'instant') {
                // first arg not an object/null
                // or behavior is auto, instant or undefined
                return true;
            }

            // first argument is an object and behavior is smooth
            if (typeof x === 'object' && x.behavior === 'smooth')
                return false;


            // throw error when behavior is not supported
            throw new TypeError('behavior not valid');
        }

        /**
         * finds scrollable parent of an element
         * @method findScrollableParent
         * @param {Node} el
         * @returns {Node} el
         */
        function findScrollableParent(el) {
            let isBody,
                hasScrollableSpace,
                hasVisibleOverflow;

            do {
                el = el.parentNode;

                // set condition variables
                isBody = el === d.body;
                hasScrollableSpace = (el.clientHeight < el.scrollHeight || el.clientWidth < el.scrollWidth);
                hasVisibleOverflow = w.getComputedStyle(el, null).overflow === 'visible';
            }
            while (!isBody && !(hasScrollableSpace && !hasVisibleOverflow));

            isBody = hasScrollableSpace = hasVisibleOverflow = null;

            return el;
        }

        /**
         * self invoked function that, given a context, steps through scrolling
         * @method step
         * @param {Object} context
         */
        function step(context) {
            // call method again on next available frame
            context.frame = w.requestAnimationFrame(step.bind(w, context));

            let time = now(),
                value,
                currentX,
                currentY,
                elapsed = (time - context.startTime) / SCROLL_TIME;

            // avoid elapsed times higher than one
            elapsed = elapsed > 1 ? 1 : elapsed;

            // apply easing to elapsed time
            value = ease(elapsed);

            currentX = context.startX + (context.x - context.startX) * value;
            currentY = context.startY + (context.y - context.startY) * value;

            context.method.call(context.scrollable, currentX, currentY);

            // return when end points have been reached
            if (currentX === context.x && currentY === context.y) {
                w.cancelAnimationFrame(context.frame);
                return;
            }
        }

        /**
         * scrolls window with a smooth behavior
         * @method smoothScroll
         * @param {Object|Node} el
         * @param {Number} x
         * @param {Number} y
         */
        function smoothScroll(el, x, y) {
            let scrollable,
                startX,
                startY,
                method,
                startTime = now(),
                frame;

            // define scroll context
            if (el === d.body) {
                scrollable = w;
                startX = w.scrollX || w.pageXOffset;
                startY = w.scrollY || w.pageYOffset;
                method = original.scroll;
            } else {
                scrollable = el;
                startX = el.scrollLeft;
                startY = el.scrollTop;
                method = scrollElement;
            }

            // cancel frame when a scroll event's happening
            if (frame)
                w.cancelAnimationFrame(frame);


            // scroll looping over a frame
            step({
                scrollable: scrollable,
                method: method,
                startTime: startTime,
                startX: startX,
                startY: startY,
                x: x,
                y: y,
                frame: frame
            });
        }

        /*
         * ORIGINAL METHODS OVERRIDES
         */

        // w.scroll and w.scrollTo
        w.scroll = w.scrollTo = function() {
            // avoid smooth behavior if not required
            if (shouldBailOut(arguments[0])) {
                original.scroll.call(
                    w,
                    arguments[0].left || arguments[0],
                    arguments[0].top || arguments[1]
                );
                return;
            }

            // doppia tilde === Math.floor() (round)
            // LET THE SMOOTHNESS BEGIN!
            smoothScroll.call(
                w,
                d.body, ~~arguments[0].left, ~~arguments[0].top
            );
        };

        // w.scrollBy
        w.scrollBy = function() {
            // avoid smooth behavior if not required
            if (shouldBailOut(arguments[0])) {
                original.scrollBy.call(
                    w,
                    arguments[0].left || arguments[0],
                    arguments[0].top || arguments[1]
                );
                return;
            }

            // LET THE SMOOTHNESS BEGIN!
            smoothScroll.call(
                w,
                d.body, ~~arguments[0].left + (w.scrollX || w.pageXOffset), ~~arguments[0].top + (w.scrollY || w.pageYOffset)
            );
        };

        // Element.prototype.scroll and Element.prototype.scrollTo
        Element.prototype.scroll = Element.prototype.scrollTo = function() {
            // avoid smooth behavior if not required
            if (shouldBailOut(arguments[0])) {
                original.elScroll.call(
                    this,
                    arguments[0].left || arguments[0],
                    arguments[0].top || arguments[1]
                );
                return;
            }

            // LET THE SMOOTHNESS BEGIN!
            smoothScroll.call(
                this,
                this,
                arguments[0].left,
                arguments[0].top
            );
        };

        // Element.prototype.scrollBy
        Element.prototype.scrollBy = function() {
            var arg0 = arguments[0];

            if (typeof arg0 === 'object') {
                this.scroll({
                    left: arg0.left + this.scrollLeft,
                    top: arg0.top + this.scrollTop,
                    behavior: arg0.behavior
                });
            } else {
                this.scroll(
                    this.scrollLeft + arg0,
                    this.scrollTop + arguments[1]
                );
            }
        };

        // Element.prototype.scrollIntoView
        Element.prototype.scrollIntoView = function() {

            // avoid smooth behavior if not required
            if (shouldBailOut(arguments[0])) {
                original.scrollIntoView.call(this, arguments[0] || true);
                return;
            }

            // LET THE SMOOTHNESS BEGIN!
            let scrollableParent = findScrollableParent(this);
            let parentRects = scrollableParent.getBoundingClientRect();
            let clientRects = this.getBoundingClientRect();



            if (scrollableParent !== d.body) {
                // reveal element inside parent
                smoothScroll.call(
                    this,
                    scrollableParent,
                    scrollableParent.scrollLeft + clientRects.left - parentRects.left,
                    scrollableParent.scrollTop + clientRects.top - parentRects.top
                );

                // reveal parent in viewport
                w.scrollBy({
                    left: parentRects.left,
                    top: parentRects.top,
                    behavior: 'smooth'
                });
            } else {

                // reveal element in viewport
                w.scrollBy({
                    left: clientRects.left,
                    top: clientRects.top,
                    behavior: 'smooth'
                });
            }
        };
    }

    if (typeof exports === 'object') {
        // commonjs
        module.exports = { polyfill: polyfill };
    } else {
        // global
        polyfill();
    }
})(window, document);

/*
 * smoothscroll polyfill - v0.3.4
 * https://iamdustan.github.io/smoothscroll
 * 2016 (c) Dustan Kasten, Jeremias Menichelli - MIT License
 */
const angular_module = require('./module');
let lastId = null;
angular_module.directive('smoothIntoView', ($timeout) => {
    return {
        link: (scope, elem, attrs) => {

            angular.element(() => {

                const delay = scope.$eval(attrs.delay) || 0;
                const ev = attrs.smoothEvent || 'click';
                const behavior = { behavior: 'smooth' };

                if (attrs.smoothBlock)
                    behavior.block = attrs.smoothBlock;

                const isEnable = () => ('undefined' === typeof scope.$eval(attrs.enableSmooth) ? true : scope.$eval(attrs.enableSmooth));


                let a = scope.$watch(() => isEnable(), enable => {

                    if (enable && lastId != attrs.smoothIntoView) {
                        elem.on(ev, () => {
                            lastId = attrs.smoothIntoView;
                            $timeout(() => {
                                document.getElementById(attrs.smoothIntoView).scrollIntoView(behavior);
                            }, delay, true);
                        });
                    } else
                        elem.off(ev);

                    a();
                    a = null;
                });

            });



        }
    };

});